@@ -254,7 +254,8 @@ _setimage(ImagingEncoderObject *encoder, PyObject *args) {
         state->ysize = y1 - y0;
     }
 
-    if (state->xsize <= 0 || state->xsize + state->xoff > im->xsize ||
+    if (state->xoff < 0 || state->xsize <= 0 ||
+        state->xsize + state->xoff > im->xsize || state->yoff < 0 ||
         state->ysize <= 0 || state->ysize + state->yoff > im->ysize) {
         PyErr_SetString(PyExc_SystemError, "tile cannot extend outside image");
         return NULL;
@@ -334,14 +335,19 @@ static PyTypeObject ImagingEncoderType = {
 /* -------------------------------------------------------------------- */
 
 int
-get_packer(ImagingEncoderObject *encoder, const char *mode, const char *rawmode) {
+get_packer(ImagingEncoderObject *encoder, const ModeID mode, const RawModeID rawmode) {
     int bits;
     ImagingShuffler pack;
 
     pack = ImagingFindPacker(mode, rawmode, &bits);
     if (!pack) {
         Py_DECREF(encoder);
-        PyErr_Format(PyExc_ValueError, "No packer found from %s to %s", mode, rawmode);
+        PyErr_Format(
+            PyExc_ValueError,
+            "No packer found from %s to %s",
+            getModeData(mode)->name,
+            getRawModeData(rawmode)->name
+        );
         return -1;
     }
 
@@ -402,11 +408,13 @@ PyObject *
 PyImaging_GifEncoderNew(PyObject *self, PyObject *args) {
     ImagingEncoderObject *encoder;
 
-    char *mode;
-    char *rawmode;
+    char *mode_name;
+    char *rawmode_name;
     Py_ssize_t bits = 8;
     Py_ssize_t interlace = 0;
-    if (!PyArg_ParseTuple(args, "ss|nn", &mode, &rawmode, &bits, &interlace)) {
+    if (!PyArg_ParseTuple(
+            args, "ss|nn", &mode_name, &rawmode_name, &bits, &interlace
+        )) {
         return NULL;
     }
 
@@ -415,6 +423,9 @@ PyImaging_GifEncoderNew(PyObject *self, PyObject *args) {
         return NULL;
     }
 
+    const ModeID mode = findModeID(mode_name);
+    const RawModeID rawmode = findRawModeID(rawmode_name);
+
     if (get_packer(encoder, mode, rawmode) < 0) {
         return NULL;
     }
@@ -435,11 +446,11 @@ PyObject *
 PyImaging_PcxEncoderNew(PyObject *self, PyObject *args) {
     ImagingEncoderObject *encoder;
 
-    char *mode;
-    char *rawmode;
+    char *mode_name;
+    char *rawmode_name;
     Py_ssize_t bits = 8;
 
-    if (!PyArg_ParseTuple(args, "ss|n", &mode, &rawmode, &bits)) {
+    if (!PyArg_ParseTuple(args, "ss|n", &mode_name, &rawmode_name, &bits)) {
         return NULL;
     }
 
@@ -448,6 +459,9 @@ PyImaging_PcxEncoderNew(PyObject *self, PyObject *args) {
         return NULL;
     }
 
+    const ModeID mode = findModeID(mode_name);
+    const RawModeID rawmode = findRawModeID(rawmode_name);
+
     if (get_packer(encoder, mode, rawmode) < 0) {
         return NULL;
     }
@@ -465,12 +479,12 @@ PyObject *
 PyImaging_RawEncoderNew(PyObject *self, PyObject *args) {
     ImagingEncoderObject *encoder;
 
-    char *mode;
-    char *rawmode;
+    char *mode_name;
+    char *rawmode_name;
     Py_ssize_t stride = 0;
     Py_ssize_t ystep = 1;
 
-    if (!PyArg_ParseTuple(args, "ss|nn", &mode, &rawmode, &stride, &ystep)) {
+    if (!PyArg_ParseTuple(args, "ss|nn", &mode_name, &rawmode_name, &stride, &ystep)) {
         return NULL;
     }
 
@@ -479,6 +493,9 @@ PyImaging_RawEncoderNew(PyObject *self, PyObject *args) {
         return NULL;
     }
 
+    const ModeID mode = findModeID(mode_name);
+    const RawModeID rawmode = findRawModeID(rawmode_name);
+
     if (get_packer(encoder, mode, rawmode) < 0) {
         return NULL;
     }
@@ -499,11 +516,11 @@ PyObject *
 PyImaging_TgaRleEncoderNew(PyObject *self, PyObject *args) {
     ImagingEncoderObject *encoder;
 
-    char *mode;
-    char *rawmode;
+    char *mode_name;
+    char *rawmode_name;
     Py_ssize_t ystep = 1;
 
-    if (!PyArg_ParseTuple(args, "ss|n", &mode, &rawmode, &ystep)) {
+    if (!PyArg_ParseTuple(args, "ss|n", &mode_name, &rawmode_name, &ystep)) {
         return NULL;
     }
 
@@ -512,6 +529,9 @@ PyImaging_TgaRleEncoderNew(PyObject *self, PyObject *args) {
         return NULL;
     }
 
+    const ModeID mode = findModeID(mode_name);
+    const RawModeID rawmode = findRawModeID(rawmode_name);
+
     if (get_packer(encoder, mode, rawmode) < 0) {
         return NULL;
     }
@@ -536,7 +556,7 @@ PyImaging_XbmEncoderNew(PyObject *self, PyObject *args) {
         return NULL;
     }
 
-    if (get_packer(encoder, "1", "1;R") < 0) {
+    if (get_packer(encoder, IMAGING_MODE_1, IMAGING_RAWMODE_1_R) < 0) {
         return NULL;
     }
 
@@ -557,8 +577,8 @@ PyObject *
 PyImaging_ZipEncoderNew(PyObject *self, PyObject *args) {
     ImagingEncoderObject *encoder;
 
-    char *mode;
-    char *rawmode;
+    char *mode_name;
+    char *rawmode_name;
     Py_ssize_t optimize = 0;
     Py_ssize_t compress_level = -1;
     Py_ssize_t compress_type = -1;
@@ -567,8 +587,8 @@ PyImaging_ZipEncoderNew(PyObject *self, PyObject *args) {
     if (!PyArg_ParseTuple(
             args,
             "ss|nnny#",
-            &mode,
-            &rawmode,
+            &mode_name,
+            &rawmode_name,
             &optimize,
             &compress_level,
             &compress_type,
@@ -597,6 +617,9 @@ PyImaging_ZipEncoderNew(PyObject *self, PyObject *args) {
         return NULL;
     }
 
+    const ModeID mode = findModeID(mode_name);
+    const RawModeID rawmode = findRawModeID(rawmode_name);
+
     if (get_packer(encoder, mode, rawmode) < 0) {
         free(dictionary);
         return NULL;
@@ -605,7 +628,7 @@ PyImaging_ZipEncoderNew(PyObject *self, PyObject *args) {
     encoder->encode = ImagingZipEncode;
     encoder->cleanup = ImagingZipEncodeCleanup;
 
-    if (rawmode[0] == 'P') {
+    if (rawmode == IMAGING_RAWMODE_P || rawmode == IMAGING_RAWMODE_PA) {
         /* disable filtering */
         ((ZIPSTATE *)encoder->state.context)->mode = ZIP_PNG_PALETTE;
     }
@@ -634,8 +657,8 @@ PyObject *
 PyImaging_LibTiffEncoderNew(PyObject *self, PyObject *args) {
     ImagingEncoderObject *encoder;
 
-    char *mode;
-    char *rawmode;
+    char *mode_name;
+    char *rawmode_name;
     char *compname;
     char *filename;
     Py_ssize_t fp;
@@ -646,16 +669,24 @@ PyImaging_LibTiffEncoderNew(PyObject *self, PyObject *args) {
     int key_int, status, is_core_tag, is_var_length, num_core_tags, i;
     TIFFDataType type = TIFF_NOTYPE;
     // This list also exists in TiffTags.py
-    const int core_tags[] = {256,   257, 258,   259, 262, 263, 266,   269,   274,
-                             277,   278, 280,   281, 340, 341, 282,   283,   284,
-                             286,   287, 296,   297, 320, 321, 338,   32995, 32998,
-                             32996, 339, 32997, 330, 531, 530, 65537, 301,   532};
+    const int core_tags[] = {256, 257, 258,   259,   262,   263,   266,  269, 274, 277,
+                             278, 280, 281,   282,   283,   284,   286,  287, 296, 297,
+                             301, 320, 321,   330,   333,   338,   339,  340, 341, 530,
+                             531, 532, 32995, 32996, 32997, 32998, 65537};
 
     Py_ssize_t tags_size;
     PyObject *item;
 
     if (!PyArg_ParseTuple(
-            args, "sssnsOO", &mode, &rawmode, &compname, &fp, &filename, &tags, &types
+            args,
+            "sssnsOO",
+            &mode_name,
+            &rawmode_name,
+            &compname,
+            &fp,
+            &filename,
+            &tags,
+            &types
         )) {
         return NULL;
     }
@@ -693,6 +724,9 @@ PyImaging_LibTiffEncoderNew(PyObject *self, PyObject *args) {
         return NULL;
     }
 
+    const ModeID mode = findModeID(mode_name);
+    const RawModeID rawmode = findRawModeID(rawmode_name);
+
     if (get_packer(encoder, mode, rawmode) < 0) {
         return NULL;
     }
@@ -788,7 +822,8 @@ PyImaging_LibTiffEncoderNew(PyObject *self, PyObject *args) {
             }
         }
 
-        if (type == TIFF_BYTE || type == TIFF_UNDEFINED) {
+        if (type == TIFF_BYTE || type == TIFF_UNDEFINED ||
+            key_int == TIFFTAG_INKNAMES) {
             status = ImagingLibTiffSetField(
                 &encoder->state,
                 (ttag_t)key_int,
@@ -922,13 +957,25 @@ PyImaging_LibTiffEncoderNew(PyObject *self, PyObject *args) {
                     );
                     free(av);
                 }
+            } else if (type == TIFF_RATIONAL) {
+                FLOAT32 *av;
+                /* malloc check ok, calloc checks for overflow */
+                av = calloc(len, sizeof(FLOAT32));
+                if (av) {
+                    for (i = 0; i < len; i++) {
+                        av[i] = (FLOAT32)PyFloat_AsDouble(PyTuple_GetItem(value, i));
+                    }
+                    status =
+                        ImagingLibTiffSetField(&encoder->state, (ttag_t)key_int, av);
+                    free(av);
+                }
             }
         } else {
             if (type == TIFF_SHORT) {
                 status = ImagingLibTiffSetField(
                     &encoder->state, (ttag_t)key_int, (UINT16)PyLong_AsLong(value)
                 );
-            } else if (type == TIFF_LONG) {
+            } else if (type == TIFF_LONG || type == TIFF_IFD) {
                 status = ImagingLibTiffSetField(
                     &encoder->state, (ttag_t)key_int, (UINT32)PyLong_AsLong(value)
                 );
@@ -944,10 +991,6 @@ PyImaging_LibTiffEncoderNew(PyObject *self, PyObject *args) {
                 status = ImagingLibTiffSetField(
                     &encoder->state, (ttag_t)key_int, (FLOAT32)PyFloat_AsDouble(value)
                 );
-            } else if (type == TIFF_DOUBLE) {
-                status = ImagingLibTiffSetField(
-                    &encoder->state, (ttag_t)key_int, (FLOAT64)PyFloat_AsDouble(value)
-                );
             } else if (type == TIFF_SBYTE) {
                 status = ImagingLibTiffSetField(
                     &encoder->state, (ttag_t)key_int, (INT8)PyLong_AsLong(value)
@@ -956,7 +999,8 @@ PyImaging_LibTiffEncoderNew(PyObject *self, PyObject *args) {
                 status = ImagingLibTiffSetField(
                     &encoder->state, (ttag_t)key_int, PyBytes_AsString(value)
                 );
-            } else if (type == TIFF_RATIONAL) {
+            } else if (type == TIFF_DOUBLE || type == TIFF_SRATIONAL ||
+                       type == TIFF_RATIONAL) {
                 status = ImagingLibTiffSetField(
                     &encoder->state, (ttag_t)key_int, (FLOAT64)PyFloat_AsDouble(value)
                 );
@@ -1076,8 +1120,8 @@ PyObject *
 PyImaging_JpegEncoderNew(PyObject *self, PyObject *args) {
     ImagingEncoderObject *encoder;
 
-    char *mode;
-    char *rawmode;
+    char *mode_name;
+    char *rawmode_name;
     Py_ssize_t quality = 0;
     Py_ssize_t progressive = 0;
     Py_ssize_t smooth = 0;
@@ -1101,8 +1145,8 @@ PyImaging_JpegEncoderNew(PyObject *self, PyObject *args) {
     if (!PyArg_ParseTuple(
             args,
             "ss|nnnnpn(nn)nnnOz#y#y#",
-            &mode,
-            &rawmode,
+            &mode_name,
+            &rawmode_name,
             &quality,
             &progressive,
             &smooth,
@@ -1130,11 +1174,14 @@ PyImaging_JpegEncoderNew(PyObject *self, PyObject *args) {
         return NULL;
     }
 
+    const ModeID mode = findModeID(mode_name);
+    RawModeID rawmode = findRawModeID(rawmode_name);
+
     // libjpeg-turbo supports different output formats.
     // We are choosing Pillow's native format (3 color bytes + 1 padding)
     // to avoid extra conversion in Pack.c.
-    if (ImagingJpegUseJCSExtensions() && strcmp(rawmode, "RGB") == 0) {
-        rawmode = "RGBX";
+    if (ImagingJpegUseJCSExtensions() && rawmode == IMAGING_RAWMODE_RGB) {
+        rawmode = IMAGING_RAWMODE_RGBX;
     }
 
     if (get_packer(encoder, mode, rawmode) < 0) {
@@ -1192,7 +1239,7 @@ PyImaging_JpegEncoderNew(PyObject *self, PyObject *args) {
     encoder->encode = ImagingJpegEncode;
 
     JPEGENCODERSTATE *jpeg_encoder_state = (JPEGENCODERSTATE *)encoder->state.context;
-    strncpy(jpeg_encoder_state->rawmode, rawmode, 8);
+    jpeg_encoder_state->rawmode = rawmode;
     jpeg_encoder_state->keep_rgb = keep_rgb;
     jpeg_encoder_state->quality = quality;
     jpeg_encoder_state->qtables = qarrays;
